@@ -3,9 +3,13 @@
 
 #include "timeline/maintimelinemodel.h"
 
+#include "networkcontroller.h"
 #include "texthandler.h"
 
 #include <KLocalizedString>
+#include <QJsonDocument>
+#include <QNetworkReply>
+#include <QUrlQuery>
 #include <config.h>
 
 MainTimelineModel::MainTimelineModel(QObject *parent)
@@ -24,9 +28,9 @@ QString MainTimelineModel::displayName() const
     if (m_timelineName == QStringLiteral("home")) {
         return i18nc("@title", "Home");
     } else if (m_timelineName == QStringLiteral("public")) {
-        return i18nc("@title", "Local Timeline");
+        return i18nc("@title", "Local");
     } else if (m_timelineName == QStringLiteral("federated")) {
-        return i18nc("@title", "Global Timeline");
+        return i18nc("@title", "Global");
     } else if (m_timelineName == QStringLiteral("bookmarks")) {
         return i18nc("@title", "Bookmarks");
     } else if (m_timelineName == QStringLiteral("favourites")) {
@@ -35,6 +39,8 @@ QString MainTimelineModel::displayName() const
         return i18nc("@title", "Trending");
     } else if (m_timelineName == QStringLiteral("list")) {
         return m_listId;
+    } else if (m_timelineName == QStringLiteral("link")) {
+        return m_url;
     }
 
     return {};
@@ -57,6 +63,23 @@ void MainTimelineModel::setListId(const QString &id)
     fillTimeline({});
 }
 
+QString MainTimelineModel::url() const
+{
+    return m_url;
+}
+
+void MainTimelineModel::setUrl(const QString &url)
+{
+    if (m_url == url) {
+        return;
+    }
+
+    m_url = url;
+    Q_EMIT urlChanged();
+
+    fillTimeline({});
+}
+
 void MainTimelineModel::setName(const QString &timelineName)
 {
     if (timelineName == m_timelineName) {
@@ -76,14 +99,16 @@ void MainTimelineModel::fillTimeline(const QString &fromId, bool backwards)
                                         QStringLiteral("bookmarks"),
                                         QStringLiteral("favourites"),
                                         QStringLiteral("trending"),
-                                        QStringLiteral("list")};
-    static const QSet publicTimelines = {QStringLiteral("home"), QStringLiteral("public"), QStringLiteral("federated")};
+                                        QStringLiteral("list"),
+                                        QStringLiteral("link")};
+    static const QSet publicTimelines = {QStringLiteral("home"), QStringLiteral("public"), QStringLiteral("federated"), QStringLiteral("link")};
 
     const bool isHome = m_timelineName == QStringLiteral("home");
     const bool isList = m_timelineName == QStringLiteral("list");
     const bool isPublic = m_timelineName == QStringLiteral("public");
     const bool isTrending = m_timelineName == QStringLiteral("trending");
     const bool isFederated = m_timelineName == QStringLiteral("federated");
+    const bool isLink = m_timelineName == QStringLiteral("link");
 
     // Ensure we aren't trying to load without an account, loading something else, or with an invalid timeline name.
     if (!m_account || loading() || !validTimelines.contains(m_timelineName)) {
@@ -101,6 +126,11 @@ void MainTimelineModel::fillTimeline(const QString &fromId, bool backwards)
         return;
     }
 
+    // If we are trying to load a link, don't continue without knowing which one to load.
+    if (isLink && m_url.isEmpty()) {
+        return;
+    }
+
     setLoading(true);
 
     QUrl url;
@@ -140,13 +170,16 @@ void MainTimelineModel::fillTimeline(const QString &fromId, bool backwards)
         // is this really how it's supposed to work wrt read markers?
         query.addQueryItem(QStringLiteral("max_id"), fromId);
     }
+    if (isLink) {
+        query.addQueryItem(QStringLiteral("url"), m_url);
+    }
     url.setQuery(query);
 
     m_account->get(
         url,
         true,
         this,
-        [this, currentTimelineName = m_timelineName, account = m_account, backwards, isHome](QNetworkReply *reply) {
+        [this, currentTimelineName = m_timelineName, account = m_account, backwards](QNetworkReply *reply) {
             // This weird m_account != account is to protect against account switches that might happen while loading
             // Ditto for timeline name
             if (m_account != account || m_timelineName != currentTimelineName) {
@@ -154,21 +187,33 @@ void MainTimelineModel::fillTimeline(const QString &fromId, bool backwards)
                 return;
             }
 
-            const auto linkHeader = QString::fromUtf8(reply->rawHeader(QByteArrayLiteral("Link")));
-
-            m_next = TextHandler::getNextLink(linkHeader);
-            Q_EMIT atEndChanged();
-
-            m_prev = TextHandler::getPrevLink(linkHeader);
-
-            if (publicTimelines.contains(m_timelineName) && backwards) {
-                int const pos = fetchedTimeline(reply->readAll());
-                Q_EMIT repositionAt(pos);
-            } else {
-                fetchedTimeline(reply->readAll(), true);
+            // If the reply is empty, do NOT overwrite m_prev/m_next and wipe pagination. That just means the server has nothing more to give us, at the moment.
+            const auto data = reply->readAll();
+            const auto doc = QJsonDocument::fromJson(data);
+            if (doc.array().isEmpty()) {
+                setLoading(false);
+                return;
             }
 
-            fetchedTimeline(reply->readAll(), !publicTimelines.contains(m_timelineName));
+            const auto linkHeader = QString::fromUtf8(reply->rawHeader(QByteArrayLiteral("Link")));
+
+            // If we're going backwards we do NOT want to overwrite m_next if it exists.
+            // Otherwise pagination breaks and the user can't load anything further in their timeline.
+            if (!backwards || !m_next) {
+                m_next = TextHandler::getNextLink(linkHeader);
+            }
+            // Load m_prev initially, then make sure never to overwrite it if we're loading new stuff
+            if (backwards || !m_prev) {
+                m_prev = TextHandler::getPrevLink(linkHeader);
+            }
+            Q_EMIT atEndChanged();
+
+            if (publicTimelines.contains(m_timelineName) && backwards) {
+                int const pos = fetchedTimeline(data);
+                Q_EMIT repositionAt(pos);
+            } else {
+                fetchedTimeline(data, true);
+            }
 
             // hasPrevious depends not just on m_prev, but also m_timeline!
             Q_EMIT hasPreviousChanged();
@@ -176,8 +221,8 @@ void MainTimelineModel::fillTimeline(const QString &fromId, bool backwards)
             setLoading(false);
         },
         [this](const QNetworkReply *reply) {
-            Q_UNUSED(reply)
             setLoading(false);
+            Q_EMIT networkErrorOccurred(reply->errorString());
         });
 }
 
@@ -189,10 +234,19 @@ void MainTimelineModel::handleEvent(AbstractAccount::StreamingEventType eventTyp
         if (eventType == AbstractAccount::StreamingEventType::UpdateEvent && m_timelineName == QStringLiteral("home")) {
             const auto doc = QJsonDocument::fromJson(payload);
             const auto post = new Post(m_account, doc.object(), this);
-            beginInsertRows({}, 0, 0);
-            m_timeline.push_front(post);
-            endInsertRows();
-            Q_EMIT streamedPostAdded(post->originalPostId());
+
+            // Make sure we aren't adding the same post we already have
+            const auto it = std::ranges::find_if(std::as_const(m_timeline), [post](const auto &timelinePost) {
+                return post->postId() == timelinePost->postId();
+            });
+            if (it == m_timeline.cend()) {
+                beginInsertRows({}, 0, 0);
+                m_timeline.push_front(post);
+                endInsertRows();
+                Q_EMIT streamedPostAdded(post->originalPostId());
+            } else {
+                delete post;
+            }
         }
     }
 }
@@ -261,7 +315,7 @@ void MainTimelineModel::fetchLastReadId()
                 fillTimeline({});
             }
         },
-        [=](QNetworkReply *reply) {
+        [this](QNetworkReply *reply) {
             Q_UNUSED(reply);
 
             // If you failed, give up
@@ -297,6 +351,17 @@ void MainTimelineModel::updateReadMarker(const QString &postId)
     }
 }
 
+void MainTimelineModel::refresh()
+{
+    // If we have pagination data, use that to refresh. Otherwise fall back to reloading the whole thing.
+    if (m_prev) {
+        fillTimeline({}, true);
+    } else {
+        reset();
+        fillTimeline({});
+    }
+}
+
 bool MainTimelineModel::canFetchMore(const QModelIndex &parent) const
 {
     Q_UNUSED(parent)
